/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>

/* This module includes */
#include "server/dtmHeader.h"

/* structures, protos and errors of the calife library */
#include "libdtm.h"



/**********************************************************************
 *
 * params check for the p3d_poster production fct (dtm_to_p3d_poster)
 *
 **********************************************************************/

int dtm_checkP3dPosterParams (int		nblig,
				    int		nbcol,
				    double	zOffset,
				    double	zScale)
{
  if ((zOffset <= 0.0) || (zScale <= 0))
    return dtmError (ERR_DTM_SCALE);
  
  if ((nblig > DTM_MAX_LINES)
      || (nbcol > DTM_MAX_COLUMNS)
      || (nblig <= 0)
      || (nbcol <= 0))
    return dtmError (ERR_DTM_SIZE);
  
/*       || (nblig > dtm->) */
/*       || (nbcol > dtm->) */
  
  /* This last check to ensure that  nblig / 2 will be OK. Just not to tackle
     the usual #$^$# pbs */
/*   if (((nblig % 2) != 0) || ((nbcol % 2) != 0)) */
/*     return dtmError (ERR_DTM_ODD_SIZE); */
  
 return (DTM_OK);
}

/**********************************************************************
 *
 * Filling of a DTM_P3D_POSTER structure. The DTM_P3D_POSTER is centered on
 * the last robot position. TODO : it could be centered on the current robot
 * position (in case a lot of time ellapsed after the last perception ), but
 * for that we need this last robot position.
 *
 **********************************************************************/

#define DTM_POSTER_MIN_POINTS 10

int
dtm_to_p3d_poster (DTM*			dtm,
		   double		p3dNbLines,
		   double		p3dNbCols,
		   double		zOffset, 
		   double		zScale,
		   T3D*			mainToOrigin, 
		   DTM_P3D_POSTER*	dtm_p3d_poster)


{
  /* Parameters to define the window in the dtm that will be copied in the
     dtm_p3d_poster */

  int		i_robot;
  int		j_robot;	/* Robot last position in the dtm  */
  int		i_min_dtm;
  int		i_max_dtm;
  int		j_min_dtm;
  int		j_max_dtm;	/* Bounds in the dtm that define the
				   dtm_p3d_poster window (might be over the
				   DTM) */  

  int		i_min_poster;
  int		i_max_poster;
  int		j_min_poster;
  int		j_max_poster;	/* Bounds in which to loop in the dtm_p3d_poster
				 */ 

  int		delta_i;
  int		delta_j;

  /* To fill the dtm_p3d_poster */
  int		i, j;		/* Pour boucler */
  DTM_CELL*	cell;		/* Une cellule du 'DTM' */
  double	zOrigin;	/* Cf commentaire plus bas */


  /* Params validity */
  if (dtm_checkP3dPosterParams (p3dNbLines, p3dNbCols, 
				zOffset, zScale) != DTM_OK)
    return (-1);
  
  switch (dtm->stockage) {
	  case X_LIGNES:
		  i_robot = (mainToOrigin->euler.euler[3] - dtm->x_origine) / dtm->scale_x;
		  j_robot = (mainToOrigin->euler.euler[4] - dtm->y_origine) / dtm->scale_y;

		  break;
	  case X_COLONNES:
		  return dtmError (ERR_DTM_FCT_NON_IMPLEMENTEE);
		  /*       break; */
		  /*     Unreached */
	  default:
		  return dtmError (ERR_DTM_CAS_IMPOSSIBLE);
		  /*     break; */
		  /*     Unreached */
  }

  if ((i_robot < 0) || (i_robot >= dtm->nblig) 
      || (j_robot < 0) || (j_robot >= dtm->nbcol)) {
    printf ("%s: Current pos : (%.2f, %.2f)\n", 
	    __func__,
	    mainToOrigin->euler.euler[3],
	    mainToOrigin->euler.euler[4]);
    printf ("%s: Coordinates of the DTM origin : (%.2f, %.2f)\n",
	    __func__,
	    dtm->x_origine, dtm->y_origine);
    printf ("%s: Robot position in the dtm in pixels : (%d, %d)\n",
	   __func__, i_robot, j_robot);
    return dtmError (ERR_DTM_ROBOT_OUT_OF_DTM);
  }

  i_min_dtm = i_robot - (p3dNbLines / 2);
  i_max_dtm = i_robot + (p3dNbLines / 2);
  j_min_dtm = j_robot - (p3dNbCols / 2);
  j_max_dtm = j_robot + (p3dNbCols / 2);

  if (i_min_dtm < 0) {
    i_min_poster = - i_min_dtm;
    delta_i = i_min_dtm;
  }
  else {
    i_min_poster = 0;
    delta_i = i_min_dtm;
  }

  if (i_max_dtm >= dtm->nblig) {
    i_max_poster = p3dNbLines - (i_max_dtm - dtm->nblig);
  }
  else {
    i_max_poster = p3dNbLines;
  }

  if (j_min_dtm < 0) {
    j_min_poster = - j_min_dtm;
    delta_j = j_min_dtm;
  }
  else {
    j_min_poster = 0;
    delta_j = j_min_dtm;
  }

  if (j_max_dtm >= dtm->nblig) {
    j_max_poster = p3dNbCols - (j_max_dtm - dtm->nbcol);
  }
  else {
    j_max_poster = p3dNbCols;
  }

  /* Initialisation  NULL des tous les champs de la DTM_P3D_POSTER */
  memset ((DTM_P3D_POSTER*) dtm_p3d_poster, 0, sizeof (DTM_P3D_POSTER));

  /* Calcul des bornes du MNT */
  if (dtm_minmax (dtm) != DTM_OK) 
    return (-1);

  /* Ecriture de l'entte */
  dtm_p3d_poster->nbLines = p3dNbLines;
  dtm_p3d_poster->nbCols = p3dNbCols;
  
  zOrigin = - zOffset - dtm->z_min;
  dtm_p3d_poster->zOrigin = zOrigin;

  dtm_p3d_poster->xScale = dtm->scale_x;
  dtm_p3d_poster->yScale = dtm->scale_y;
  dtm_p3d_poster->zScale = (float) zScale;

  dtm_p3d_poster->xOrigin = (i_min_dtm * dtm->scale_x) + dtm->x_origine;
  dtm_p3d_poster->yOrigin = (j_min_dtm * dtm->scale_y) + dtm->y_origine;

  /* Copy of the altitude infos */
  for (i = i_min_poster; i < i_max_poster; i++) {
	  for (j = j_min_poster; j < j_max_poster; j++) {
		  cell = dtm_one_cell(dtm, i + delta_i, j + delta_j);

		  if (cell->state == EMPTY || cell->current_info.nb_points < DTM_POSTER_MIN_POINTS)
		  {
			  dtm_p3d_poster->state[i][j] = DTM_CELL_EMPTY;
			  dtm_p3d_poster->zfloat[i][j] = 0.0;
		  }
		  else 
		  {
			  if (cell->hstate == VERTICAL) {
				  dtm_p3d_poster->zfloat[i][j] = cell->current_info.z_max;
			  } else {
				  dtm_p3d_poster->zfloat[i][j] = cell->current_info.z_moyen;
			  }
			  dtm_p3d_poster->state[i][j] = DTM_CELL_FILLED;
		  }


	  }
  }

  return (DTM_OK);
}

/**********************************************************************
 *
 * Filling of a DTM_LABEL_POSTER structure. Code mostly duplicated
 * from the dtm_to_p3d_poster function
 *
 **********************************************************************/

#define DTM_POSTER_MIN_POINTS 10

int
dtm_to_label_poster (DTM* dtm, DTM_LABEL_POSTER*	dtm_label_poster)
{
  unsigned char* pu;
  DTM_CELL* cell;
  size_t i, size;

  if (dtm->nblig != dtm_label_poster->nbLines ||
	  dtm->nbcol != dtm_label_poster->nbCols)
    return dtmError (ERR_DTM_SIZE);

  dtm_label_poster->xOrigin = dtm->x_origine;
  dtm_label_poster->yOrigin = dtm->y_origine;
  dtm_label_poster->xScale = dtm->scale_x;
  dtm_label_poster->yScale = dtm->scale_y;

  cell = dtm->cells_tab;
  pu = dtm_label_poster->state;
  size = dtm->nblig * dtm->nbcol;

  for (i = 0; i < size; ++i, cell++, pu++) {
      if (cell->state == EMPTY || cell->current_info.nb_points < DTM_POSTER_MIN_POINTS)
		  *pu = DTM_NO_LABEL;
      else {
		  if (cell->hstate == VERTICAL) 
			  *pu = DTM_LABEL_OBSTACLE;
		  else 
			  *pu = DTM_LABEL_TRAVERSABLE;
      }
    }

  return (DTM_OK);
}

/**********************************************************************
 *
 * Writing of a DTM_P3D_POSTER structure
 *
 **********************************************************************/

int
write_dtm_p3d_poster (DTM_P3D_POSTER*	dtm,
		      FILE*		fichier)

{
  int		i, j;		/* Pour boucler */

  fprintf (fichier, DTM_P3D_POSTER_ASCII_FF);
  fprintf (fichier, "\n");
  fprintf (fichier, "%d %d\n", dtm->nbLines, dtm->nbCols);
  fprintf (fichier, "%f %f %f\n", 
	   dtm->xOrigin, dtm->yOrigin, dtm->zOrigin);
  fprintf (fichier, "%f %f %f\n", 
	   dtm->xScale, dtm->yScale, dtm->zScale);

  for (i = 0; i < dtm->nbLines; i++) {
    for (j = 0; j < dtm->nbCols; j++) {

      fprintf (fichier, "%d\t", dtm->state[i][j]);
      fprintf (fichier, "%f\t", dtm->zfloat[i][j]);
    }
  }

  return (DTM_OK);  

}

/**************************************************************
 *
 * Creation of a DTM poster (structure DTM_P3D_POSTER) 
 *
 **************************************************************/

DTM_P3D_POSTER*
create_dtm_p3d_poster ( )

{
  DTM_P3D_POSTER*	dtm;	/* Ce que l'on va crer */

  /* Simple allocation */
  dtm = (DTM_P3D_POSTER*) malloc (sizeof (DTM_P3D_POSTER));
  if (dtm == NULL) {
    dtmError (ERR_DTM_ALLOC_POSTER);
    return (NULL);
  }
  
  /* Initialisation  NULL des tous les champs de la DTM_P3D_POSTER */
  memset ((DTM_P3D_POSTER*) dtm, 0, sizeof (DTM_P3D_POSTER));

  return (dtm);
}

/**********************************************************************
 * Destruction of a DTM poster
 *
 * *******************************************************************/
void
destroy_dtm_p3d_poster(DTM_P3D_POSTER* poster)
{
	free(poster);
}

/**********************************************************************
 *
 * Reading of a DTM_P3D_POSTER structure
 *
 **********************************************************************/

DTM_P3D_POSTER*
read_dtm_p3d_poster (char*		fileName)

{
  int		i, j;		/* Pour boucler */
  FILE*		fichier;

  char          entete[256];

  DTM_P3D_POSTER*	dtm;

  /* Cration du DTM_P3D_POSTER */
  dtm = create_dtm_p3d_poster ();
  if (dtm == NULL) {
    dtmError (ERR_DTM_ALLOC_POSTER);
    return (NULL);
  }

  fichier = fopen (fileName, "r");
  if (fichier == NULL) {
    dtmError(ERR_DTM_OPEN_FILE);
	goto fail;
  }

  fscanf (fichier, "%s", entete);
  if (strcmp (entete, DTM_P3D_POSTER_ASCII_FF) != 0) {
    dtmError (ERR_DTM_FILEFORMAT);
	goto fail2;
  }

  if (fscanf (fichier, "%d %d\n", &(dtm->nbLines), &(dtm->nbCols)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
	goto fail2;
  }
  if (fscanf (fichier, "%f %f %f\n", 
	      &(dtm->xOrigin), &(dtm->yOrigin), &(dtm->zOrigin)) != 3) {
    dtmError (ERR_DTM_FILEFORMAT);
	goto fail2;
  }
  if (fscanf (fichier, "%f %f %f\n", 
	      &(dtm->xScale), &(dtm->yScale), &(dtm->zScale)) != 3) {
    dtmError (ERR_DTM_FILEFORMAT);
	goto fail2;
  }

  for (i = 0; i < dtm->nbLines; i++) {
    for (j = 0; j < dtm->nbCols; j++) {
	  int state;
	  float z;

	  if (fscanf (fichier, "%d\t%f\n", &state, &z) != 2) {
		  dtmError (ERR_DTM_FILEFORMAT);
		  goto fail2;
      }
	  dtm->state[i][j] = state;
	  dtm->zfloat[i][j] = z;
    }
  }
  
  fclose (fichier);
  return (dtm);  

fail2:
  fclose(fichier);
fail:
  destroy_dtm_p3d_poster(dtm);
  return NULL;
}

